/*
 * shell.c
 *
 * subroutines for shelling
 *
 * Ken Stevens, 1996
 */

#include "player.h"
#include "nat.h"
#include "sect.h"
#include "ship.h"
#include "land.h"
#include "shell.h"
#include "options.h"
#include "optlist.h"
#include "mission.h"
#include "nsc.h"
#include "file.h"
#include "var.h"
#include "news.h"
#include "damage.h"

static int maximum_distance _PROTO((struct shelist *vlist, int x, int y));
static int sct_canfire _PROTO((struct shelist *vlist, struct sctstr *asectp, int verbose));
static int interdict_range _PROTO((struct shelist *vlist, struct genitem *gp, int verbose));
static int shp_canfire _PROTO((struct shelist *vlist, struct shpstr *ashipp, int verbose));
static int lnd_canfire _PROTO((struct shelist *vlist, struct lndstr *alandp, int verbose));
static int num_shots _PROTO((struct genitem *gp));
static int can_hit _PROTO((struct genitem *vitem, struct genitem *aitem, double range));

/* Initialize a new shelist struct */

shell_init_shelist(listp)
	struct shelist *listp;
{
	bzero(listp, sizeof(struct shelist));
	initque(&listp->queue);
	initque(&listp->targets.queue);
}

/* Find things to fire from */

shell_find_fire(vlist, victim, att_string, attacker_ef_type, alist, att_cnum, trigger, verbose)
	struct	shelist *vlist;		/* victim items */
	natid	victim;			/* victim cnum */
	s_char *att_string; 		/* player argument for fire */
	int	attacker_ef_type;	/* attacker's ef type */
	struct	shelist *alist;		/* attacker list of shelling items */
	int	att_cnum;		/* attacker country number */
	int	trigger;		/* who called this function? */
	int	verbose;		/* should we report bad items? */
{
	struct	nstr_item nbst; /* list of firing items */
	struct	nstr_sect nstr; /* list of firing sectors */
	struct	sctstr asect;	/* attacking sector */
	struct	shpstr aship;	/* attacking ship */
	struct	lndstr aland;	/* attacking land */
	int	ammo_cost;
	char	hate_victim[MAXNOC]; /* countries which will fire on victim */
	natid	cnum;
	int	radius;
	coord	cx,cy;

	if (QEMPTY(&vlist->queue))
		return;

	cx = ((struct shelist *)vlist->queue.q_forw)->item->x;
	cy = ((struct shelist *)vlist->queue.q_forw)->item->y;
	radius = maximum_distance(vlist, cx, cy) +
		(int)(SHELL_MAX_FIRE_RANGE * fire_range_factor + 0.9999);
	bzero(hate_victim, sizeof(hate_victim));
	if (att_cnum == SHELL_HOSTILE) {
		for (cnum = 1; cnum < MAXNOC; ++cnum)
			if (getrel(getnatp(cnum), victim) < NEUTRAL)
				hate_victim[cnum] = 1;
	} else {
		hate_victim[att_cnum] = 1;
		if (trigger != SHELL_FIRE_COMMAND &&
		    trigger != SHELL_TORP_COMMAND)
			for (cnum = 1; cnum < MAXNOC; ++cnum)
				if (getrel(getnatp(cnum), victim) == AT_WAR &&
				    getrel(getnatp(cnum), att_cnum) == ALLIED)
					hate_victim[cnum] = 1;
	}

	if (opt_FORTFIRE &&
	    trigger != SHELL_UNIT_INTERDICTION &&
	    trigger != SHELL_TORP_COMMAND &&
	    (attacker_ef_type == EF_SECTOR ||
	     attacker_ef_type == EF_BAD) &&
	    snxtsct(&nstr, att_string)) {
		if (nbst.sel == NS_ALL)
			snxtsct_dist(&nstr, cx, cy, radius);
		while(nxtsct(&nstr, &asect)) {
			if (!hate_victim[asect.sct_own])
				continue;
			if (asect.sct_type != SCT_FORTR)
				continue;
			if (!(ammo_cost = sct_canfire(vlist, &asect, verbose)))
				continue;
			insque(shell_create_item((struct genitem *)&asect, ammo_cost, 0, trigger), &alist->queue);
		}
	}
	if ((attacker_ef_type == EF_SHIP ||
	     attacker_ef_type == EF_BAD) &&
	    snxtitem(&nbst, EF_SHIP, att_string)) {
		if (nbst.sel == NS_ALL)
			snxtitem_dist(&nbst, EF_SHIP, cx, cy, radius);
		while(nxtitem(&nbst, &aship)) {
			if (!hate_victim[aship.shp_own])
				continue;
			if (!mchr[aship.shp_type].m_frnge)
				continue;
			if ((trigger == SHELL_SHIP_INTERDICTION ||
			     trigger == SHELL_UNIT_INTERDICTION) &&
			    aship.shp_mission != MI_INTERDICT)
				continue;
			if (!(ammo_cost = shp_canfire(vlist, &aship, verbose)))
				continue;
			if (trigger == SHELL_TORP_COMMAND) {
				if (!(mchr[aship.shp_type].m_flags & M_TORP))
					continue;
				if (aship.shp_mobil <= 0) {
					if (verbose)
						mpr(aship.shp_own, "%s doesn't have enough mobility to torpedo!\n", prship(&aship));
					continue;
				}
			}
			if (aship.shp_mission == MI_INTERDICT &&
			    !interdict_range(vlist, (struct genitem *)&aship, verbose))
				continue;
			insque(shell_create_item((struct genitem *)&aship, ammo_cost, 0, trigger), &alist->queue);
		}
	}
	if ((attacker_ef_type == EF_LAND ||
	     attacker_ef_type == EF_BAD) &&
	    trigger != SHELL_TORP_COMMAND &&
	    snxtitem(&nbst, EF_LAND, att_string)) {
		if (nbst.sel == NS_ALL)
			snxtitem_dist(&nbst, EF_LAND, cx, cy, radius);
		while(nxtitem(&nbst, &aland)) {
			if (!hate_victim[aland.lnd_own])
				continue;
			if (!lchr[aland.lnd_type].l_frg)
				continue;
			if ((trigger == SHELL_SHIP_INTERDICTION ||
			     trigger == SHELL_UNIT_INTERDICTION) &&
			    aland.lnd_mission != MI_INTERDICT)
				continue;
			if (aland.lnd_ship >= 0) {
				if (verbose)
					mpr(att_cnum, "%s can't fire because it's on a ship!\n", prland(&aland));
				continue;
			}
			if (!(ammo_cost = lnd_canfire(vlist, &aland, verbose)))
				continue;
			if (aland.lnd_mission == MI_INTERDICT &&
			    !interdict_range(vlist, (struct genitem *)&aland, verbose))
				continue;

			insque(shell_create_item((struct genitem *)&aland, ammo_cost, 0, trigger), &alist->queue);;
		}
	}
}

static double
sct_range(asectp)
	struct sctstr *asectp;
{
	return asectp->sct_effic / 60 + tfactfire(asectp->sct_own, 7.0);
}

/* The number of shells spent to fire */

static int
fire_ammo_cost(gp)
	struct	genitem *gp;
{
 	return max(ldround(num_shots(gp)/2.0, 1), 1);
}

static int
maximum_distance(vlist, x, y)
	struct	shelist *vlist;		/* victim items */
	int x, y;
{
	register struct qelem *qp;
	register struct shelist *vlp;
	
	int distance = 0;

	for (qp = vlist->queue.q_back; qp != &vlist->queue; qp = qp->q_back) {
		vlp = (struct shelist *)qp;
		distance = max(distance, mapdist(vlp->item->x, vlp->item->y, x, y));
	}

	return distance;
}

static int
minimum_distance(vlist, x, y)
	struct	shelist *vlist;		/* victim items */
{
	register struct qelem *qp;
	register struct shelist *vlp;
	int distance = 999;

	for (qp = vlist->queue.q_back; qp != &vlist->queue; qp = qp->q_back) {
		vlp = (struct shelist *)qp;
		distance = min(distance,
			       mapdist(vlp->item->x, vlp->item->y, x, y));
	}

	return distance;
}

static int
sct_draw_supply(asectp, ammo_cost)
	struct sctstr *asectp;
	int ammo_cost;
{
	int	vec[I_MAX+1];

	getsect(asectp->sct_x, asectp->sct_y, asectp);
	getvec(VT_ITEM, vec, (s_char *)asectp, EF_SECTOR);
	vec[I_SHELL] += supply_commod(asectp->sct_own, asectp->sct_x,
				      asectp->sct_y, I_SHELL, ammo_cost);
	putvec(VT_ITEM, vec, asectp, EF_SECTOR);
	putsect(asectp);
}

static int
sct_canfire(vlist, asectp, verbose)
	struct	shelist *vlist;		/* victim items */	
	struct sctstr *asectp;
	int verbose;
{
	int	vec[I_MAX+1];
	int	mdist;
	double	srange;
	int	ammo_cost;

	if (!getsect(asectp->sct_x, asectp->sct_y, asectp) ||
	    !asectp->sct_own)
		return 0;
	mdist = minimum_distance(vlist, asectp->sct_x, asectp->sct_y);
	srange = sct_range(asectp);
	ammo_cost = fire_ammo_cost(asectp);
	if (mdist > srange + 0.9) {
		if (verbose && mdist < srange + 2.9)
			mpr(asectp->sct_own, "Sector %s can't reach target (range is %.2f and distance to target is %d)\n",
			    ownxy(asectp), srange, mdist);		
		return 0;
	}

	if (asectp->sct_effic < ((u_char)FORTEFF)) {
		mpr(asectp->sct_own, "Fort at %s not efficient enough to fire!  (must be %d%%)\n",
		   ownxy(asectp), FORTEFF);
		return 0;
	}	

	getvec(VT_ITEM, vec, (s_char *)asectp, EF_SECTOR);
	if (!vec[I_GUN]) {
		if (verbose)
			mpr(asectp->sct_own, "Not enough guns in sector %s!  (need at least 1)\n",
			   ownxy(asectp));
		return 0;
	}
	if (vec[I_SHELL] < ammo_cost) {
		sct_draw_supply(asectp, ammo_cost);
	}
	if (vec[I_SHELL] < ammo_cost) {
		if (verbose)
			mpr(asectp->sct_own, "Not enough shells in sector %s!  (need at least %d)\n",
			   ownxy(asectp), ammo_cost);
		return 0;
	}
	if (vec[I_MILIT] < 5) {
		if (verbose)
			mpr(asectp->sct_own, "Not enough military in sector %s!  (need at least 5)\n",
			   ownxy(asectp));
		return 0;
	}
	return ammo_cost;
}

double
shell_shp_range(ashipp)
	struct shpstr *ashipp;
{
	return techfactfire(ashipp->shp_tech, mchr[ashipp->shp_type].m_frnge / 2.0);
}

static int
shp_draw_supply(ashipp, ammo_cost)
	struct shpstr *ashipp;
	int ammo_cost;
{
	int	vec[I_MAX+1];

	getship(ashipp->shp_uid, ashipp);
	getvec(VT_ITEM, vec, (s_char *)ashipp, EF_SHIP);
	vec[I_SHELL] += supply_commod(ashipp->shp_own, ashipp->shp_x,
				      ashipp->shp_y, I_SHELL, ammo_cost);
	putvec(VT_ITEM, vec, ashipp, EF_SHIP);
	putship(ashipp->shp_uid, ashipp);
}

s_char *
pritem(itemp, who)
	struct	genitem *itemp;
	natid	who;
{
	switch(itemp->ef_type) {
	case EF_SECTOR:
		return prbuf("%s %s", dchr[itemp->type].d_name, xyas(itemp->x, itemp->y, who));
	case EF_SHIP:
		return prship((struct shpstr *)itemp);
	case EF_LAND:
		return prland((struct lndstr *)itemp);
	}
}

static int
interdict_range(vlist, gp, verbose)
	struct	shelist *vlist;
	struct	genitem *gp;
	int	verbose;
{
	int	mdist;
	double	srange;

	mdist = minimum_distance(vlist, gp->opx, gp->opy);
	if (mdist > gp->radius) {
		if (verbose && mdist < srange + 2.9)
			mpr(gp->own, "target is outside of mission operation area of %s (op area is @%d,%d:%d and distance to target is %d)\n",
			    pritem(gp), gp->opx, gp->opy, gp->radius, mdist);
		return 0;
	}
	return 1;
}

static int
shp_canfire(vlist, ashipp, verbose)
	struct	shelist *vlist;		/* victim items */	
	struct shpstr *ashipp;
	int verbose;
{
	int	vec[I_MAX+1];
	int	mdist;
	double	srange;
	int	ammo_cost;

	if (!getship(ashipp->shp_uid, ashipp) ||
	    !ashipp->shp_own)
		return 0;

	mdist = minimum_distance(vlist, ashipp->shp_x, ashipp->shp_y);
	srange = shell_shp_range(ashipp);
	ammo_cost = fire_ammo_cost(ashipp);
	if (mdist > srange + 0.9) {
		if (verbose && mdist < srange + 2.9)
			mpr(ashipp->shp_own, "%s can't reach target (range is %.2f and distance to target is %d)\n",
			    prship(ashipp), srange, mdist);		
		return 0;
	}

	if (ashipp->shp_effic < 60) {
		mpr(ashipp->shp_own, "%s is crippled and cannot fire!  (must be %d%%)\n",
		   prship(ashipp), 60);
		return 0;
	}	

	getvec(VT_ITEM, vec, (s_char *)ashipp, EF_SHIP);
	if (!vec[I_GUN]) {
		if (verbose)
			mpr(ashipp->shp_own, "Not enough guns on %s!  (need at least 1)\n",
			   prship(ashipp));
		return 0;
	}
	if (vec[I_SHELL] < ammo_cost) {
		shp_draw_supply(ashipp, ammo_cost);
	}
	if (vec[I_SHELL] < ammo_cost) {
		if (verbose)
			mpr(ashipp->shp_own, "Not enough shells on %s!  (need at least %d)\n",
			   prship(ashipp), ammo_cost);
		return 0;
	}
	if (vec[I_MILIT] < 1) {
		if (verbose)
			mpr(ashipp->shp_own, "Not enough military on %s!  (need at least 1)\n",
			   prship(ashipp));
		return 0;
	}

	return ammo_cost;
}

double
shell_lnd_range(alandp)
	struct lndstr *alandp;
{
	return techfactfire(alandp->lnd_tech, (double)lchr[alandp->lnd_type].l_frg);
}

static int
lnd_draw_supply(alandp, ammo_cost)
	struct lndstr *alandp;
	int ammo_cost;
{
	resupply_commod(alandp, I_SHELL);
}

static int
lnd_canfire(vlist, alandp, verbose)
	struct	shelist *vlist;		/* victim items */	
	struct lndstr *alandp;
	int verbose;
{
	int	vec[I_MAX+1];
	int	mdist;
	double	srange;
	int	ammo_cost;

	if (!getland(alandp->lnd_uid, alandp) ||
	    !alandp->lnd_own)
		return 0;
	mdist = minimum_distance(vlist, alandp->lnd_x, alandp->lnd_y);
	srange = shell_lnd_range(alandp);
	ammo_cost = fire_ammo_cost(alandp);
	if (mdist > srange + 0.9) {
		if (verbose && mdist < srange + 2.9)
			mpr(alandp->lnd_own, "%s can't reach target (range is %.2f and distance to target is %d)\n",
			    prland(alandp), srange, mdist);		
		return 0;
	}

	if (alandp->lnd_effic < LAND_MINFIREEFF) {
		mpr(alandp->lnd_own, "%s is crippled and cannot fire!  (must be %d%%)\n",
		   prland(alandp), LAND_MINFIREEFF);
		return 0;
	}	

	getvec(VT_ITEM, vec, (s_char *)alandp, EF_LAND);
	if (vec[I_SHELL] < ammo_cost) {
		lnd_draw_supply(alandp, ammo_cost);
		putland(alandp->lnd_uid, alandp);
	}
	getvec(VT_ITEM, vec, (s_char *)alandp, EF_LAND);
	if (vec[I_SHELL] < ammo_cost) {
		if (verbose)
			mpr(alandp->lnd_own, "%s -- not enough shells!  (need at least %d)\n",
			   prland(alandp), ammo_cost);
		return 0;
	}

	return ammo_cost;
}

static int
num_shots(gp)
	struct genitem *gp;
{
	int guns = SHELL_MAX_GUNS;

	switch (gp->ef_type) {
	case EF_LAND:
		guns = min(guns, lchr[gp->type].l_gun);
		break;
	case EF_SHIP:
		guns = min(guns, mchr[gp->type].m_glim);
		/* fall through to default: */
	default:
		guns = min(guns, getvar(V_GUN, (s_char *)gp, gp->ef_type));
	}

	return guns;
}

static
getitem(itemp)
	struct genitem *itemp;
{
	switch (itemp->ef_type) {
	case EF_SECTOR:
		return getsect(itemp->x, itemp->y, (struct sctstr *)itemp);
	case EF_SHIP:
		return getship(itemp->uid, (struct shpstr *)itemp);
	case EF_LAND:
		return getland(itemp->uid, (struct lndstr *)itemp);
	}
	return 0;
}

static
putitem(itemp)
	struct genitem *itemp;
{
	switch (itemp->ef_type) {
	case EF_SECTOR:
		putsect((struct sctstr *)itemp);
		break;
	case EF_SHIP:
		putship(itemp->uid, (struct shpstr *)itemp);
		break;
	case EF_LAND:
		putland(itemp->uid, (struct lndstr *)itemp);
	}
}


int
roundrange(r)
	double r;
{
	double f;

	f=r-((int)r);
	if (chance(f))
		return(((int)r)+1);
	else
		return((int)r);
}

static double
do_fire(alp)
	struct shelist *alp;
{
	int	shell;
	int	i, shots, damage = 0;
	extern int torpedo_damage;

	if (!getitem(alp->item))
		return 0;
	shell = getvar(V_SHELL, (s_char *)alp->item, alp->item->ef_type);
	shell -= alp->ammo_cost;
	if (shell < 0) {
		/* impossible */
		putvar(V_SHELL, 0, (s_char *)alp->item, alp->item->ef_type);
		putitem(alp->item);
		return 0;
	}
	putvar(V_SHELL, shell, (s_char *)alp->item, alp->item->ef_type);
	if (shell < alp->ammo_cost)
		alp->draw_supply(alp->item, alp->ammo_cost);
	putitem(alp->item);

	shots = num_shots(alp->item);
	if (shots)
		alp->rounds_left = getvar(V_SHELL, (s_char *)alp->item, alp->item->ef_type) * 2 / shots;
	else
		alp->rounds_left = -1;

	if (alp->item->ef_type == EF_SHIP) {
		if (alp->use_torp || !opt_NOMOBCOST) {
			if (alp->item->mobil <= 0) /* if numtimes > 1 */
				return 0;
			alp->item->mobil = max(alp->item->mobil - 15, -100);
			putitem(alp->item);
		}
	}

	if (roundrange(alp->fire_range) < alp->distance) {
		mpr(alp->start_own, "%s misses target\n", pritem(alp->item, alp->start_own));
		return 0;
	}
	if (alp->use_torp) {
		return TORP_DAMAGE();
	}
	for (i = 0; i <shots; ++i) {
		damage += 10 + random() % 6;
	}
	return damage * 0.01 * alp->item->effic;
}

static
set_shelist(slist, ammo_cost)
	struct shelist *slist;
	int ammo_cost;
{
	slist->ammo_cost = ammo_cost;
	slist->start_eff = slist->item->effic;
	slist->start_own = slist->item->own;
	slist->fire_range = slist->range(slist->item);
}

struct shelist *
shell_create_item(itemp, ammo_cost, lp, trigger)
	struct	genitem *itemp;
	int	ammo_cost;
	struct	shelist *lp;
	int	trigger;
{
	struct	shelist *new_item;
	int	allocated = 0;

	if (!(new_item = lp))
		new_item = (struct shelist *)malloc(sizeof(struct shelist));
	shell_init_shelist(new_item);

	switch (itemp->ef_type) {
	case EF_SECTOR:
		if (lp)
			new_item->item = itemp;
		else {
			new_item->item = (struct genitem *)malloc(sizeof(struct sctstr));
			bcopy(itemp, new_item->item, sizeof(struct sctstr));
			allocated = 1;
		}
		new_item->canfire = sct_canfire;
		new_item->range = sct_range;
		new_item->draw_supply = sct_draw_supply;
		break;
	case EF_SHIP:
		if (lp)
			new_item->item = itemp;
		else {
			new_item->item = (struct genitem *)malloc(sizeof(struct shpstr));
			bcopy(itemp, new_item->item, sizeof(struct shpstr));
			allocated = 1;
		}
		new_item->canfire = shp_canfire;
		new_item->range = shell_shp_range;
		new_item->draw_supply = shp_draw_supply;
		if (is_sub(new_item->item) || trigger == SHELL_TORP_COMMAND)
			new_item->use_torp = 1;
		break;
	case EF_LAND:
		if (lp)
			new_item->item = itemp;
		else {
			new_item->item = (struct genitem *)malloc(sizeof(struct lndstr));
			bcopy(itemp, new_item->item, sizeof(struct lndstr));
			allocated = 1;
		}
		new_item->canfire = lnd_canfire;
		new_item->range = shell_lnd_range;
		new_item->draw_supply = lnd_draw_supply;
		break;
	default:
		return 0;
	}

	set_shelist(new_item, ammo_cost);
	new_item->allocated = allocated;

	return new_item;
}

/* Is vitem a valid thing we can shoot at? */

shell_valid_target(vitem, att_cnum, verbose)
	struct genitem *vitem;
	natid	att_cnum;
	int	verbose;
{
	switch (vitem->ef_type) {
	case EF_SECTOR:
		if (!getsect(vitem->x, vitem->y, vitem)) {
			if (verbose)
				mpr(att_cnum, "No such sector exists!\n");
			return 0;
		}
		if (vitem->type == SCT_SANCT) {
			if (verbose)
				mpr(att_cnum, "You can't shell the sanctuary at %s!\n",
				   xyas(vitem->x, vitem->y, player->cnum));
			return 0;
		}
		if (vitem->type == SCT_WATER) {
			if (verbose)
				mpr(att_cnum, "You must specify a ship in sector %s!",
				   xyas(vitem->x, vitem->y, player->cnum));
			return 0;
		}
		break;
	case EF_SHIP:
		if (!getship(vitem->uid, vitem) ||
		    !vitem->own) {
			if (verbose)
				mpr(att_cnum, "No such ship exists!\n");
			return 0;
		}
		break;
	case EF_LAND:
		if (!getland(vitem->uid, vitem) ||
		    !vitem->own) {
			if (verbose)
				mpr(att_cnum, "No such land unit exists!\n");
			return 0;
		}
		break;
	default:
		return 0;
	}

	return 1;
}

static
free_list(list)
	struct qelem *list;
{
	register struct qelem *qp, *next;

	for (qp = list->q_forw; qp != list; qp = next) {
		next = qp->q_forw;
		remque(qp);
		free((s_char *)qp);
	}
}

shell_select_targets(vlist, alist, verbose)
	struct shelist *vlist;
	struct shelist *alist;
	int verbose;
{
	register struct qelem *aqp;
	register struct shelist *alp;
	register struct qelem *vqp;
	register struct shelist *vlp;
	struct targlist *targp;

	for (aqp = alist->queue.q_back; aqp != &alist->queue; aqp = aqp->q_back) {
		alp = (struct shelist *)aqp;
		alp->distance = 999;
		for (vqp = vlist->queue.q_back; vqp != &vlist->queue; vqp = vqp->q_back) {
			vlp = (struct shelist *)vqp;
			if (!shell_valid_target(vlp->item, alp->start_own, verbose))
				continue;
			if (can_hit(vlp->item, alp->item, alp->fire_range)) {
				targp = (struct targlist *)malloc(sizeof(struct targlist));
				targp->target = vlp;
				insque(targp, &alp->targets.queue);
				targp->distance = mapdist(vlp->item->x, vlp->item->y, alp->item->x, alp->item->y);
				if (targp->distance < alp->distance)
					alp->distance = targp->distance;
			}
		}
		if (QEMPTY(&alp->targets.queue)) {
			remque(alp);
			free((s_char *)alp);
		}
	}
	
}

static int
can_hit(vitem, aitem, range)
	struct genitem *vitem;
	struct genitem *aitem;
	double range;
{
	if (mapdist(vitem->x, vitem->y, aitem->x, aitem->y) > range + 0.9)
		return 0;

	switch (vitem->ef_type) {
	case EF_LAND:
	case EF_SECTOR:
		return !is_sub(aitem);
	case EF_SHIP:
		return !is_sub(vitem) ||
			(aitem->ef_type == EF_SHIP &&
			 (mchr[aitem->type].m_flags & (M_DCH | M_SUBT)));
	default:
		return 0;
	}

	return 1;
}

shell_reset(vlist, alist, verbose)
	struct shelist *vlist;
	struct shelist *alist;
	int verbose;
{
	register struct qelem *qp, *next;
	register struct shelist *alp;
	int ammo_cost;

	for (qp = alist->queue.q_back; qp != &alist->queue; qp = next) {
		alp = (struct shelist *)qp;
		next = qp->q_back;
		if (!(ammo_cost = alp->canfire(vlist, alp->item, verbose))) {
			remque(qp);
			free(qp);
			continue;
		}
		alp->damage_delivered = 0;
		alp->damage_received = 0.0;
		free_list(&alp->targets.queue);
		set_shelist(alp, ammo_cost);
	}
}
	
static int
she_count(tlist)
	struct targlist *tlist;
{
	struct qelem *qp;
	int count = 0;
	
	for (qp = tlist->queue.q_forw; qp != &tlist->queue; qp = qp->q_forw)
		++count;

	return count;
}

static
itemdamage(itemp, damage)
	struct genitem *itemp;
	int damage;
{
	switch(itemp->ef_type) {
	case EF_SECTOR:
		sectdamage((struct sctstr *)itemp, damage);
		break;
	case EF_SHIP:
		shipdamage((struct shpstr *)itemp, damage);
		break;
	case EF_LAND:
		landdamage((struct lndstr *)itemp, damage);
		break;
	}
}

shell_damage(vlist)
	struct shelist *vlist;
{
	register struct qelem *vqp;
	register struct shelist *vlp;

	for (vqp = vlist->queue.q_back; vqp != &vlist->queue; vqp = vqp->q_back) {
		vlp = (struct shelist *)vqp;
		getitem(vlp->item);
		itemdamage(vlp->item, (int)vlp->damage_received);
		putitem(vlp->item);
	}
}

static double
damage_distance(damage, fire_range, distance)
	double	damage;
	double	fire_range;
	int	distance;
{
	int effective_range;

	if (fire_range > 1.0)
		effective_range = roundrange(fire_range - 1.0);
	else
		effective_range = 0.0;

	return distance > effective_range ? damage / 2 : damage;
}

int
shell_fire(vlist, alist)
	struct shelist *vlist;
	struct shelist *alist;
{
	register struct qelem *aqp, *anext;
	register struct shelist *alp;
	register struct qelem *tqp;
	register struct targlist *tlp;
	int count;
	double damage_each;
	double total_damage = 0.0;

	for (aqp = alist->queue.q_back; aqp != &alist->queue; aqp = anext) {
		alp = (struct shelist *)aqp;
		anext = aqp->q_back;
		if (!(count = she_count(&alp->targets))) {
			/* not possible */
			remque(aqp);
			free((s_char *)aqp);
			continue;
		}
		
		total_damage += alp->damage_delivered = do_fire(alp);
		damage_each = alp->damage_delivered / count;
		for (tqp = alp->targets.queue.q_back; tqp != &alp->targets.queue; tqp = tqp->q_back) {
			tlp = (struct targlist *)tqp;
			tlp->target->damage_received += damage_distance(damage_each, alp->fire_range, tlp->distance);
		}
	}
	return (int)total_damage;
}

shell_free_list(list)
	struct shelist *list;
{
	struct qelem *qp, *next;
	struct shelist *lp;

	for (qp = list->queue.q_back; qp != &list->queue; qp = next) {
		lp = (struct shelist *)qp;
		next = qp->q_back;
		if (lp->allocated)
			free(lp->item);
		free_list(&lp->targets.queue);
		remque(qp);
		free(qp);
	}
}

static
print_sfrom(s, cnum, vlist)
	s_char *s;
	natid cnum;
	struct shelist *vlist;
{
	register struct qelem *vqp;
	register struct shelist *vlp;
	s_char buf[1024];
	s_char xy[16];

	strcpy(buf, s);
	for (vqp = vlist->queue.q_back; vqp != &vlist->queue; vqp = vqp->q_back) {
		vlp = (struct shelist *)vqp;
		sprintf(xy, " %s ", xyas(vlp->item->x, vlp->item->y, cnum));
		if (strstr(buf, xy))
			continue;
		if (strlen(buf) + strlen(xy) > 80)
			strncat(buf, "\n", 1024);
		    strncat(buf, xy, 1024);
	}
	mpr(cnum, "%s\n", buf);
}

static
only_subs(alist)
	struct shelist *alist;
{
	register struct qelem *aqp;
	register struct shelist *alp;
	int	informed = 0;

	for (aqp = alist->queue.q_back; aqp != &alist->queue; aqp = aqp->q_back) {
		alp = (struct shelist *)aqp;
		if (!is_sub(alp->item))
			return 0;
	}
	return 1;
}
static
only_ships(alist)
	struct shelist *alist;
{
	register struct qelem *aqp;
	register struct shelist *alp;
	int	informed = 0;

	for (aqp = alist->queue.q_back; aqp != &alist->queue; aqp = aqp->q_back) {
		alp = (struct shelist *)aqp;
		if (alp->item->ef_type != EF_SHIP)
			return 0;
	}
	return 1;
}

shell_report_damage_received(vlist, alist, att_cnum)
	struct shelist *vlist;
	struct shelist *alist;
	int att_cnum;
{
	register struct qelem *vqp;
	register struct shelist *vlp;
	int	informed = 0;

	for (vqp = vlist->queue.q_back; vqp != &vlist->queue; vqp = vqp->q_back) {
		vlp = (struct shelist *)vqp;
		if (!vlp->damage_received)
			continue;
		if (!vlp->start_own)
			continue;
		if (!informed) {
			informed = 1;
			if (att_cnum == SHELL_HOSTILE) {
				print_sfrom("Interdiction!  Hostile shells incoming from ", vlp->start_own, alist);
			} else if (only_subs(alist)) {
				print_sfrom("Incoming torpedos from ", vlp->start_own, alist);
			} else {
				print_sfrom(prbuf("Incoming %s (#%d) shells from ", cname(att_cnum), player->cnum), vlp->start_own, alist);
			}
		}
		mpr(vlp->start_own,
		    "%24.24s %3d%% eff takes %4d damage.  New eff is %3d%%\n",
		    pritem(vlp->item, vlp->start_own), vlp->start_eff,
		    (int)vlp->damage_received, vlp->item->effic);
	}
}

shell_report_damage_delivered(vlist, alist, att_cnum, vitem, victim, trigger)
	struct shelist *vlist;
	struct shelist *alist;
	int att_cnum;
	struct genitem *vitem;
	natid victim;
	int trigger;		/* what triggered the fire? */
{
	struct qelem *qp;
	struct shelist *lp;
	int damage_delivered = 0;
	int damage_received = 0;
	s_char fired[MAXNOC];
	int i;
	int victim_only_ships = only_ships(vlist);

	if (QEMPTY(&alist->queue))
		return;
	
	bzero(fired, sizeof(fired));
	if (vitem) {
		fired[att_cnum] = 1;
	} else {
		for (qp = alist->queue.q_back; qp != &alist->queue; qp = qp->q_back) {
			lp = (struct shelist *)qp;
			fired[lp->start_own] = 1;
		}
	}		

	for (i = 0; i < MAXNOC; ++i) {
		if (!fired[i])
			continue;

		switch (trigger) {
		case SHELL_RETURN_FIRE:
			if (i == att_cnum)
				mpr(i, "Return fire...\n");
			else
				mpr(i, "Assisting %s (#%d) in return fire against %s (#%d)...\n",
				    cname(att_cnum), att_cnum, cname(victim), victim);
			break;
		case SHELL_SHIP_INTERDICTION:
			print_sfrom(prbuf("Interdicting %s (#%d) ships in ",cname(victim), victim), i, vlist);
			break;
		case SHELL_UNIT_INTERDICTION:
			print_sfrom(prbuf("Interdicting %s (#%d) land units in \n",
					  cname(victim), victim), i, vlist);
			break;
		case SHELL_COMBAT_OSUPPORT:
			if (i == att_cnum)
				mpr(i, "Support fire...\n");
			else
				mpr(i, "Supporting %s's (#%d) attack on %s (#%d)...\n",
				    cname(att_cnum), att_cnum, cname(victim), victim);
			break;
		case SHELL_COMBAT_DSUPPORT:
			if (i == att_cnum)
				mpr(i, "Support fire...\n");
			else
				mpr(i, "Supporting %s's (#%d) defense against attack by %s (#%d)...\n",
				    cname(att_cnum), att_cnum, cname(victim), victim);
			break;
		}
		mpr(i, "                                             ");
		if (vitem)
			mpr(i, "      return start  new");
		mpr(i, " rounds\n");
		mpr(i, "  sect   type                    damage range");
		if (vitem)
			mpr(i, " dist damage   eff  eff");
		mpr(i, "   left\n");
		
		for (qp = alist->queue.q_back; qp != &alist->queue; qp = qp->q_back) {
			lp = (struct shelist *)qp;
			if (lp->start_own != i)
				continue;
			switch (trigger) {
			case SHELL_FIRE_COMMAND:
			case SHELL_TORP_COMMAND:
				if (victim_only_ships) {
					if (is_sub(lp->item))
						nreport(victim, N_TORP_SHIP, victim, 1);
					else if (lp->use_torp)
						nreport(victim, N_SHIP_TORP, victim, 1);
					else
						nreport(i, N_SHP_SHELL, victim, 1);
				} else {
					nreport(i, N_SCT_SHELL, victim, 1);
				}
				break;
			case SHELL_RETURN_FIRE:
				nreport(i, N_FIRE_BACK, victim, 1);
				break;
			case SHELL_SHIP_INTERDICTION:
				nreport(i, N_SHP_SHELL, victim, 1);
				break;
			case SHELL_COMBAT_OSUPPORT:
			case SHELL_COMBAT_DSUPPORT:
				switch (lp->item->ef_type) {
				case EF_SECTOR:
					nreport(i, N_FIRE_F_ATTACK, victim, 1);
					break;
				case EF_SHIP:
					nreport(i, N_FIRE_S_ATTACK, victim, 1);
					break;
				case EF_LAND:
					nreport(i, N_FIRE_L_ATTACK, victim, 1);
					break;
				}
				break;
			}
			mpr(i, "%4d,%-4d", xrel(getnatp(i),lp->item->x), yrel(getnatp(i), lp->item->y));
			mpr(i, "%-24.24s", (lp->item->ef_type == EF_SECTOR)?"fort":pritem(lp->item));
			if (lp->damage_delivered)
				mpr(i, "%6d ", (int)lp->damage_delivered);
			else
				mpr(i, "%6s ", "");
			damage_delivered += (int)lp->damage_delivered;
			mpr(i, " %1.2f ", lp->fire_range);
			if (vitem) {
				mpr(i, "%4d ", mapdist(vitem->x, vitem->y, lp->item->x, lp->item->y));
				if (lp->damage_received) {
					mpr(i, "%6d ", (int)lp->damage_received);
					damage_received += (int)lp->damage_received;
					mpr(i, "%4d%% ", lp->start_eff);
					mpr(i, "%3d%%", lp->item->effic);
				} else {
					mpr(i, "%6s %4s  %3s ", "", "", "");
				}
			}
			mpr(i, " %5d\n", lp->rounds_left);
		}
		mpr(i, "         TOTAL               %10d ", damage_delivered);
		if (vitem)
			mpr(i, "%17d", damage_received);
		mpr(i, "\n");
	}
}
